import torch
import copy
import logging
import transformers
import functools
from typing import Union
import tiktoken

DEVICE = 'cuda' if torch.cuda.is_available() else 'cpu'


class MultipleChoiceModel:
    """
    Wrapper around huggingface base LLMS and openai chat models for multiple-choice prompting, implementing a `get_scores` method.

    Its main role is to extract the computed logits to obtain probabilities for the multiple choices.
    """

    def __init__(self, model_name, labels, prompt_start_for_cache=None, openai_client=None):

        if openai_client:
            tokenizer = tiktoken.encoding_for_model(model_name)    # for o1, mind https://github.com/openai/tiktoken/issues/367
            labels = [l.strip() for l in labels]    # Hmmmm strip spaces
            labels_tokenized = [tokenizer.encode(label) for label in labels]
            label_ids = [label_tokenized[0] for label_tokenized in labels_tokenized]
            # TODO: For o1 model, logprobs not supported; so consider disabling the logprobs and just getting the output directly?
            self.get_probs = functools.partial(self.get_multiple_choice_prob_openai, model_name=model_name, client=openai_client, labels=labels, label_ids=label_ids)

            problematic_labels = [label for label, label_tokenized in zip(labels, labels_tokenized) if len(label_tokenized) != 1]
            if problematic_labels:
                raise ValueError(f'Some of the choice labels do not map onto single tokens for your selected model: {", ".join(problematic_labels)}')
            logging.info(f'Using label ids: {label_ids}')

        else:  # assuming local transformers model
            tokenizer = transformers.AutoTokenizer.from_pretrained(model_name, clean_up_tokenization_spaces=False)
            labels_tokenized = [tokenizer.encode(label, add_special_tokens=False) for label in labels]
            model = transformers.AutoModelForCausalLM.from_pretrained(model_name).to(DEVICE)
            model.eval()

            cached_common_start = create_cache(model, tokenizer, prompt_start_for_cache) if prompt_start_for_cache else None
            self.get_probs = functools.partial(self.get_multiple_choice_prob,
                                               model=model,
                                               tokenizer=tokenizer,
                                               labels_tokenized=labels_tokenized,
                                               cache=cached_common_start)


    def get_scores(self, prompt: Union[str, list[str]]) -> list[float]:
        pass    # Hmmm... Overwritten by init.

    @staticmethod
    def get_multiple_choice_prob(prompt: Union[str, list[dict]], model, tokenizer, labels_tokenized: list[int], cache=None) -> list[float]:
        """
        Feeds prompt into a local transformers model, and obtains the probabilities of the different multiple-choice
        labels.

        Because the labels can in principle be multi-token, it takes a bit of work to get their probabilities, with iterative
        prompting based on growing prefixes, and multiplying the resulting conditional probabilities.

        When labels are single tokens, this function simply applies the model once and get the logits directly.
        """

        # TODO: Generalize and expose this function, maybe in a more general LLM utils module?

        if isinstance(prompt, str):
            prompt_encoded = tokenizer(prompt, return_tensors='pt').to(DEVICE)
        else:   # openai-style messages
            prompt_encoded = tokenizer.apply_chat_template(prompt, return_tensors="pt", add_generation_prompt=True, return_dict=True).to(DEVICE)

        choice_probabilities = [1 for _ in labels_tokenized]  # to be iteratively multiplied (conditional probabilities)
        for prefix, label_ids, next_token_ids in get_label_prefixes(tuple(tuple(t) for t in labels_tokenized)):
            prompt_encoded_plus_prefix = torch.cat(
                (prompt_encoded['input_ids'], torch.tensor([prefix], dtype=int).to(DEVICE)),
                dim=-1
            )
            attention_mask_plus_prefix = torch.cat(
                (prompt_encoded['attention_mask'], torch.tensor([[1] * len(prefix)], dtype=int).to(DEVICE)),
                dim=-1
            )
            model_output = model.generate(
                input_ids=prompt_encoded_plus_prefix,
                attention_mask=attention_mask_plus_prefix,  # only to avoid a warning
                pad_token_id=tokenizer.eos_token_id,
                output_logits=True,
                return_dict_in_generate=True,
                do_sample=False,
                max_new_tokens=1,
                past_key_values=copy.deepcopy(cache) if cache else None,
                temperature=None,  # only to avoid a warning
                top_p=None
            )

            next_token_ids_unique = list(set(next_token_ids))
            logits = model_output.logits[0]  # [0] is first (and only) generated token
            logits_to_use = logits[:, next_token_ids_unique]
            probs_to_use = torch.nn.functional.softmax(logits_to_use, dim=-1)[0].tolist()   # [0] is first (and only) batch
            token_id_to_prob = dict(zip(next_token_ids_unique, probs_to_use))

            for label_id, next_token_id in zip(label_ids, next_token_ids):
                choice_probabilities[label_id] *= token_id_to_prob[next_token_id]

        return choice_probabilities


    @staticmethod
    def get_multiple_choice_prob_openai(prompt: list[dict], client, model_name, labels, label_ids: list[int]) -> list[float]:

        LOGIT_BIAS = 10

        completion = client.chat.completions.create(
            model=model_name,
            messages=prompt,
            logprobs=True,
            top_logprobs=10,
            logit_bias={l: LOGIT_BIAS for l in label_ids},
            max_completion_tokens=10,
        )

        label_logprobs = {}
        for logprob_dict in completion.choices[0].logprobs.content[0].top_logprobs:
            token, logprob = logprob_dict.token, logprob_dict.logprob
            if token in labels:
                label_logprobs[token] = logprob - LOGIT_BIAS

        newline = "\n"
        logging.info(f'{prompt[-1]["content"].replace(newline, "/")} -> {completion.choices[0].message.content}')

        label_logprobs_tensor = torch.tensor([label_logprobs.get(label, -90) - LOGIT_BIAS for label in labels])
        probs = torch.nn.functional.softmax(label_logprobs_tensor, dim=-1)
        return probs.tolist()


def create_cache(model, tokenizer, common_start: str) -> transformers.DynamicCache:
    with torch.no_grad():
        if isinstance(common_start, str):
            inputs = tokenizer(common_start, return_tensors="pt").to(DEVICE)
        else:
            inputs = tokenizer.apply_chat_template(common_start, return_tensors="pt", return_dict=True).to(DEVICE)

        ## In case multiple beams/sequences, also expand the cache (https://github.com/huggingface/transformers/pull/27576)
        # inputs = inputs.expand(num_beams, *inputs.shape[1:])

        cache = model(input_ids=inputs['input_ids'],
                      attention_mask=inputs['attention_mask'], # only to avoid a warning
                      past_key_values=transformers.DynamicCache(),
                      pad_token_id=tokenizer.eos_token_id).past_key_values
    return cache


@functools.cache
def get_label_prefixes(labels_tokenized: tuple[tuple[int]]) -> list[tuple[tuple[int], list[int], list[int]]]:
    """
    # TODO: Generalize.

    For computing probabilities of multi-token labels (e.g., the two-token string '-1' for negative sentiment),
    beam search isn't quite suitable. So we manually implement beam search, but specifically one beam for each
    possible prefix. To that end, this helper function computes all 'shared prefixes' and their next tokens
    (and index of the corresponding labels).

    Yields 3-tuples for each (shared) prefix and a list of the label_ids that share it, and their respective next_token_ids.

    >>> get_label_prefixes(((2, ), (3, )))
    [((), (0, 1), (2, 3))]
    >>> get_label_prefixes(((1, 2), (1, 3)))
    [((1), (0, 1), (2, 3)), ((), (0, 1), (1, 1)]
    >>> get_label_prefixes(((1, 2, 3), (1, 4, 5), (0, 2), (1, 5)))
    [((1, 2), [0], [3]), ((1, 5), [], []), ((1, 4), [1], [5]), ((1,), [0, 1, 3], [2, 4, 5]), ((0, 2), [], []), ((0,), [2], [2]), ((), [0, 1, 2, 3], [1, 1, 0, 1])]
    """

    max_length = max(len(t) for t in labels_tokenized)
    unique_prefixes = {l[:n] for n in range(max_length) for l in labels_tokenized}
    result = []
    for prefix in unique_prefixes:
        label_ids = []
        next_token_ids = []
        for i, l in enumerate(labels_tokenized):
            if l[:len(prefix)] == prefix and len(l) > len(prefix):
                label_ids.append(i)
                next_token_ids.append(l[len(prefix)])
        result.append((prefix, label_ids, next_token_ids))
    return result


import pandas
import sys
import argparse
import logging

import math

import random
import numpy as np
from choicellm import N_DECIMALS

"""
From the comparative results: 

target_id,comparison_id,position,target,score,choices,scores

to:

target_id,target,score,entropy,method

"""

def main():

    logging.basicConfig(level=logging.INFO, format='')
    argparser = argparse.ArgumentParser()
    argparser.add_argument('file', nargs='?', default=sys.stdin)
    argparser.add_argument('--onlyscore', action='store_true', help='whether to output only the scores; otherwise full csv.')   # TODO: re-implement
    argparser.add_argument('--n_positions', required=False, type=int, default=None, help='In case comparison data contains all positions, randomly sample one position per comparison')
    argparser.add_argument('--n_comparisons', required=False, type=int, default=None, help='Randomly sample only this many comparisons per word.')
    argparser.add_argument('--scale', required=False, type=str, help='start,end of the scale, to which to map the scores', default='1,5')
    argparser.add_argument('--seed', required=False, type=int, default=None, help='seed, for sampling --n_comparisons or --n_positions only')

    args = argparser.parse_args()
    scale_start, scale_end = (int(i) for i in args.scale.split(','))

    converters = {
        'choices': lambda x: tuple(x.split(';')),
        'probs': lambda x: tuple(float(s) for s in x.split(';')),
    }

    df = pandas.read_csv(args.file, index_col=None, converters=converters)

    n_choices = len(df.iloc[0]['choices'])  # hmmm, expose as command line arg?

    if args.n_positions is not None or args.n_comparisons is not None:
        args.seed = args.seed or random.randint(0, 999999)
        logging.info(f'Seed: {args.seed}')
        np.random.seed(args.seed)

    if args.n_positions is not None:
        df = df.groupby(['target_id', 'comparison_id']).sample(n=args.n_positions)

    if args.n_comparisons is not None:
        df = df.set_index(['target_id', 'comparison_id']).pivot(columns=['position']).groupby('target_id').sample(args.n_comparisons).stack(level=1).reset_index()

    df['entropy'] = [-sum(prob * math.log2(max(prob, 0.00000001)) for prob in scores) for scores in df['probs']]    # Hmmmmm

    df_agg_per_word = df.groupby(['target_id', 'target']).agg({'prob': 'mean', 'entropy': 'mean'})

    df_agg_per_word['rating'] = (df_agg_per_word['prob'] ** (1 / (n_choices - 1))) * (scale_end - scale_start) + scale_start
    # rating based on normalized prob: if prob(winning 1 vs 3) = 0.25 = prob(winning 1 vs 1)**3, so prob(winning 1 vs 1) = 0.25 ** (1/3)

    df_agg_per_word['rating'] = df_agg_per_word['rating'].round(N_DECIMALS)
    df_agg_per_word['entropy'] = df_agg_per_word['entropy'].round(N_DECIMALS)

    df_agg_per_word.reset_index()[['target_id', 'target', 'rating', 'entropy']].to_csv(sys.stdout, index=None)



if __name__ == '__main__':
    main()


import random
from typing import Literal, Union, Optional
import itertools
import copy
import json
import logging
import argparse


# TODO: Drastic refactoring needed
# TODO: Make fields in prompt template optional, i.e., not used if not present;
# TODO: Add optional 'mid point' field to prompt template.

class PromptTemplate:

    """
    Wrapper around plain string prompt templates, and openai-style message lists, both exposing the method `format`.

    Justification:
    - Instantiating prompt templates from prompt info (e.g., .json file) is slightly different in case of scalar, comparative and categorical prompting.
    - I want openai-style 'messages' (list of dicts) to behave the same, on the outside, as a formatable string.
    """

    def __init__(self, mode: Literal['scalar', 'comparative', 'categorical'], is_chat: bool, *args, n_choices: int = None, **kwargs):

        self.mode = mode
        self.is_chat = is_chat
        if n_choices:
            self.n_choices = n_choices  # meh
            kwargs['n_choices'] = n_choices  # meh

        # TODO: Call the following in load_from_json:
        system_prompt, examples, prompt = (
            self.init_for_scalar if mode == 'scalar'
            else self.init_for_comparative if mode == 'comparative'
            else self.init_for_categorical
        )(*args, **kwargs)

        if mode == 'comparative':
            self.labels_for_logits = kwargs['labels'][:kwargs['n_choices']]  # TODO labels_for_logits is the model's business...
        elif mode == 'categorical':
            self.labels_for_logits = kwargs['labels'][:len(kwargs['categories'])]
            self.categories = kwargs['categories']
        else:
            self.scale = kwargs['scale']
            self.labels_for_logits = [str(i) for i in kwargs['scale']]
        if not self.is_chat:  # meh...
            self.labels_for_logits = [' ' + l for l in self.labels_for_logits]

        if is_chat:
            self.prompt_format = [
                {"role": "developer", "content": system_prompt},
                *itertools.chain(*([{"role": "user", "content": example}, {"role": "assistant", "content": response}] for example, response in examples)),
                {"role": "user", "content": prompt}
            ]
            def format(*args, **kwargs):
                messages = copy.deepcopy(self.prompt_format)
                messages[-1]['content'] = messages[-1]['content'].format(*args, **kwargs)
                return messages
            self.prompt_start_for_cache = self.prompt_format[:-1]
            self.format = format
        else:
            prompt_parts = [system_prompt] + [x + ' ' + r for x, r in examples] + [prompt]
            self.prompt_start_for_cache = '\n\n'.join(prompt_parts[:-1])
            self.prompt_format = '\n\n'.join(prompt_parts)
            self.format = self.prompt_format.format

    def format(self, *args, **kwargs) -> Union[str, list[dict]]:
        pass    # to be overwritten by init; is that 'normal'?

    def __str__(self):
        if isinstance(self.prompt_format, str):
            return self.prompt_format
        else:
            return json.dumps(self.prompt_format, indent=2)

    @classmethod
    def from_json(cls, file):

        prompt_info = json.load(file)  # TODO: JSON validation, incl. labels for categorical/comparative must be strings
        prompt_info.pop('_comment', None)

        if "chat" not in prompt_info:
            logging.warning(
                f'WARNING: The prompt .json file does not specify whether to use chat-style prompting; assuming "chat": false')
            prompt_kwargs['chat'] = False

        prompt_info['is_chat'] = prompt_info.pop('chat')

        mode = prompt_info['mode']
        is_chat = prompt_info['mode']

        if mode in ('categorical', 'comparative') and 'labels' not in prompt_info:
            # TODO: In future, allow using categories/items themselves as labels; for now, just ABCD...
            prompt_info['labels'] = list(string.ascii_uppercase)

        return cls(**prompt_info)


    @staticmethod
    def init_for_scalar(system_prompt: str, prompt_format: str, examples: list[dict], scale: list[int | float]) -> tuple[str, list[tuple[str, str]], str]:

        if all(isinstance(n, int) for n in scale):
            rating_as_int = True
        else:
            scale = [float(n) for n in scale]

        labels = [str(number) for number in scale]
        scale_min, scale_max = scale[0], scale[-1]
        system_prompt = system_prompt.format(scale=', '.join(labels), scale_min=scale_min, scale_max=scale_max)
        label_hint = f' (on scale {scale_min}-{scale_max})'

        examples_list = []
        n = 0   # in case no examples
        for n, example in enumerate(examples, start=1):
            rating_scaled = example['target_value'] * (scale_max - scale_min) + scale_min
            rating_scaled = int(rating_scaled) if rating_as_int else float(rating_scaled)
            examples_list.append((
                prompt_format.format(n=n, item=example['item'], label_hint=label_hint),
                str(rating_scaled),
            ))
        prompt = prompt_format.format(n=n+1, item='{}', label_hint=label_hint)

        return system_prompt, examples_list, prompt

    @staticmethod
    def init_for_categorical(system_prompt: str, prompt_format: str, examples: list[dict], categories: dict, labels: Optional[list[str]] = None) -> tuple[str, list[tuple[str, str]], str]:
        if labels:
            categories_full = '\n'.join(f'{l}. {c}: {d}' for l, (c, d) in zip(labels, categories.items()))
        else:
            categories_full = '\n'.join(f'- {c}: {d}' for c, d in categories.items())
        category_names = list(categories.keys())
        system_prompt = system_prompt.format(categories=categories_full)
        examples_list = []
        n = 0

        label_hint = f' (choose from {"/".join(labels)})' if labels else ''

        for n, example in enumerate(examples, start=1):
            examples_list.append((
                prompt_format.format(n=n, item=example['item'], label_hint=label_hint),
                f"{labels[example['target_index']]} ({category_names[example['target_index']]})"
            ))

        prompt = prompt_format.format(n=n+1, item='{}', label_hint=label_hint)

        return system_prompt, examples_list, prompt

    @staticmethod
    def init_for_comparative(system_prompt: str, prompt_format: str, n_choices: int, examples: list[dict], labels: list[str] = None) -> tuple[str, list[tuple[str, str]], str]:

        if labels:
            labels = labels[:n_choices]

        def make_choices_str(choices: list[str], labels: Optional[list[str]] = None) -> str:
            if labels:
                return '\n'.join((f'{label}. {choice}' for label, choice in zip(labels, choices)))
            else:
                return '\n'.join(f'- {choice}' for choice in choices)

        label_hint = f' (choose from {"/".join(labels)})' if labels else ''
        system_prompt = system_prompt
        n = 0
        examples_list = []
        for n, example in enumerate(examples, start=1):

            if example['target_index'] >= n_choices:
                target = example['options'][example['target_index']]
                example['target_index'] = random.randint(0, n_choices-1)
                example['options'][example['target_index']] = target
            example['options'] = example['options'][:n_choices]

            examples_list.append((
                prompt_format.format(n=n, choices=make_choices_str(example['options'], labels), label_hint=label_hint),
                labels[example['target_index']]
            ))
        prompt = prompt_format.format(n=n+1, choices=make_choices_str(['{}'] * n_choices, labels), label_hint=label_hint)
        return system_prompt, examples_list, prompt


TEMPLATE_SCALAR = {
    "mode": "scalar",
    "chat": False,
    "system_prompt": "# Concrete vs. abstract\n\nSome words and phrases are more concrete, some are more abstract. "
                     "We can indicate how concrete a given word or phrase is, as a rating on a scale {scale}, "
                     "with {scale_min} very abstract, and {scale_max} very concrete.",
    "prompt_format": "## Example {n}.\n\nWord/phrase: {item}\n\nConcreteness rating{label_hint}:",
    "scale": [1, 2, 3, 4, 5],
    "examples": [
        {"item": "essentialness", "target_value": 0.0},
        {"item": "frangipane", "target_value": 1.0},
        {"item": "although", "target_value": 0.0},
        {"item": "blackbird", "target_value": 1.0},
        {"item": "bat", "target_value": 1.0},
        {"item": "hope", "target_value": 0.0},
    ],
    "_comment": "In the examples, \"target_value\" is the target rating as a float between [0, 1]. This will be automatically mapped to whichever scale is used. This facilitates trying different scales with the same examples."
}

TEMPLATE_SCALAR_CHAT = TEMPLATE_SCALAR.copy()
TEMPLATE_SCALAR_CHAT.update({
    'chat': True,
    'system_prompt': 'Some words and phrases are more concrete, some are more abstract. You are a helpful assistant, '
                     'who is an expert on rating how *concrete* a given word or phrase is, as a rating on a scale '
                     '{scale}, with {scale_min} very abstract, and {scale_max} very concrete.',
    'prompt_format': '## Question {n}.\n\nWord/phrase: {item}\n\nHow concrete is this {label_hint}?',
})

TEMPLATE_COMPARATIVE = {
    "mode": "comparative",
    "chat": False,
    "system_prompt": "# Concrete vs. abstract\n\nSome words and phrases are more concrete, some are more abstract. "
                     "We can often tell which word or phrase, from a given set, is the _most concrete_ one.",
    "prompt_format": "## Example {n}.\n\n{choices}\n\nThe most concrete is{label_hint}:",  # TODO: Insert "choose from labels" only if choices are not single words?
    "n_choices": 4,
    "labels": ["A", "B", "C", "D"],
    "examples": [
        {"options": ["essentialness", "simulation", "bat", "living"], "target_index": 2},
        {"options": ["blackbird", "high", "cause", "although"], "target_index": 0},
        {"options": ["signature", "frangipane", "hope", "simulation"], "target_index": 1},
    ],
    "_comment": "If \"labels\" is omitted, will use the options themselves as responses (recommended only if the options are single words). Under \"examples\", \"target_index\" is always the integer index of the correct choice in the list, 0-based."
}

TEMPLATE_COMPARATIVE_CHAT = TEMPLATE_COMPARATIVE.copy()
TEMPLATE_COMPARATIVE_CHAT.update({
    'chat': True,
    'system_prompt': 'Some words and phrases are more concrete, some are more abstract. You are a helpful assistant, '
                     'who is an expert on deciding which of several words or phrases is the _most concrete_.',
    'prompt_format': '## Question {n}.\n\n{choices}\n\nWhich of these is the most concrete{label_hint}?',
})

TEMPLATE_CATEGORICAL = {
    "mode": "categorical",
    "chat": False,
    "system_prompt": "# Concrete vs. abstract\n\nSome words and phrases are more concrete, some are more abstract. "
                     "We distinguish the following categories:\n\n{categories}",
    "prompt_format": "## Example {n}.\n\nWord/phrase: {item}\n\nThis word/phrase fits best in category{label_hint}:",
    "labels": ["A", "B", "C"],
    "categories": {
        "concrete": "the word refers to something actual, concrete, empirical",
        "neutral": "the word is neither abstract nor concrete",
        "abstract": "the word refers to something conceptual, intangible, theoretical or vague"
    },
    "examples": [
        {"item": "essentialness", "target_index": 0},
        {"item": "frangipane", "target_index": 2},
        {"item": "although", "target_index": 0},
        {"item": "blackbird", "target_index": 2},
        {"item": "bat", "target_index": 2},
        {"item": "hope", "target_index": 0}
    ],
    "_comment": "If \"labels\" is omitted, will use the options themselves as responses (recommended only if the category names are single words). Under \"examples\", \"target_index\" is always the integer index of the correct choice in the list, 0-based."
}

TEMPLATE_CATEGORICAL_CHAT = TEMPLATE_CATEGORICAL.copy()
TEMPLATE_CATEGORICAL_CHAT.update({
    'chat': True,
    'system_prompt': 'Some words and phrases are more concrete, some are more abstract. You are a helpful assistant, '
                     'who is an expert on categorizing words and phrases into one of three categories:\n\n{categories}',
    'prompt_format': '## Question {n}.\n\n{item}\n\nIn which category does this word/phrase fit best{label_hint}?',
})


def main():

    argparser = argparse.ArgumentParser('Auxiliary command to generate prompt template .json files.')
    argparser.add_argument('--chat', action='store_true', help='Whether to prompt the model like a chat/instruct model; otherwise plain text generation.')
    group = argparser.add_mutually_exclusive_group(required=True)
    group.add_argument('--comparative', action='store_true', help='Generate a prompt template for choicellm --mode comparative.')
    group.add_argument('--scalar', action='store_true', help='Generate a prompt template for choicellm --mode scalar.')
    group.add_argument('--categorical', action='store_true', help='Generate a prompt template for choicellm --mode categorical.')

    args = argparser.parse_args()

    logging.basicConfig(level=logging.INFO, format='')

    mode = 'comparative' if args.comparative else 'scalar' if args.scalar else 'categorical'

    template = globals()[f'TEMPLATE_{mode.upper()}{"_CHAT" if args.chat else ""}']
    template_as_json = json.dumps(template, indent=2)

    logging.info(f'Creating a JSON-format prompt template for {mode}, {"chat-style" if args.chat else "plain text generation"} prompting. '
                 f'Save this to a .json file, and modify to suit your needs.')

    print(template_as_json)


if __name__ == '__main__':
    main()



import argparse
import csv
import sys
import itertools
import random
import logging
from typing import Generator, Union, Iterable
import os
from openai import OpenAI
import functools
import io

from prompttemplate import PromptTemplate
from multichoicemodel import MultipleChoiceModel


N_DECIMALS = 6


def main():

    argparser = argparse.ArgumentParser()
    argparser.add_argument('file', nargs='?', type=argparse.FileType('r'), default=sys.stdin,
                           help='Plaintext file with one word/phrase per line. '
                                'Set the --newlines flag to allow the inclusion of "\\n" to indicate newlines within items. '
                                'Set the --csv flag to allow inputs in .csv format with one column and no header. '
                                'For comparative mode, the csv input can also have n_choices columns.')

    argparser.add_argument('--prompt', required=True, type=argparse.FileType('r'), default=None, help='.json file containing the prompt template, few-shot examples, etc. To generate a suitable template, first use the auxiliary command choicellm-template and adapt the template to your needs.')
    argparser.add_argument('--model', required=False, type=str, default="unsloth/Llama-3.2-1B", help='Currently supports base models via transformers, and chat models through OpenAI (specify --openai in that case).')
    argparser.add_argument('--openai', action='store_true', help='Set this flag for models through the OpenAI API; otherwise --model is assumed to be a local model via huggingface transformers.')

    input_options = argparser.add_argument_group('flags to change input format')
    input_options_excl = input_options.add_mutually_exclusive_group()
    input_options_excl.add_argument('--csv', action='store_true', help='Set this flag if your input file is in .csv format, either a single column to allow newlines etc., or multiple columns (optionally) for comparative mode.')
    input_options_excl.add_argument('--newlines', action='store_true', help='If you want your input file to include newlines, either use full, proper .csv format (and the --csv flag), or use this flag to simply replace occurrences of "\\n" in the input file by proper newlines.')

    # TODO low-priority implement this, maybe replacing --all_positions
    # argparser.add_argument('--n_orders', type=int, help='[not implemented yet] Whether to randomize the order of the categories, and if so, how often; -1 means all orders.', default=None)

    # If --mode comparative:
    comparative_group = argparser.add_argument_group('options for comparative mode (only if comparisons not predetermined in .csv input)')
    comparative_group.add_argument('--compare_to', required=False, type=argparse.FileType('r'), default=None, help='File containing the words to compare against. Default is the main file argument itself. If --csv, then --compare_to must be a csv file as well.')
    comparative_group.add_argument('--compare_deterministic', required=False, action='store_true', help='To make selection of alternatives deterministic; if --compare_to is given (and no overlap with items), this will result in the exact same comparisons per item.')
    comparative_group.add_argument('--n_comparisons', required=False, type=int, default=100, help='Comparisons per stimulus.')
    comparative_group.add_argument('--all_positions', action='store_true', help='Whether to average over all positions; only if comparative.')
    comparative_group.add_argument('--seed', required=False, type=int, default=None, help='Seed to use for shuffling/sampling of comparisons. By default a random seed is picked. If --compare_deterministic, this seed is reused for each item\'s comparisons.')

    args = argparser.parse_args()

    logging.basicConfig(level=logging.INFO, format='')
    if args.seed is None:
        args.seed = random.randint(0, 99999)
        random.seed(args.seed)
        logging.info(f'{args.seed=}')

    inputs, is_multicolumn = read_inputs(args.file, args.csv, args.newlines)

    prompt_template = PromptTemplate.from_json(args.prompt)

    if is_multicolumn and prompt_template.mode != 'comparative':
        raise ValueError('Input .csv appears to have multiple columns, which is an option only for \'comparative\' mode.')
    if args.model == argparser.get_default('model'):
        logging.warning(f'WARNING: Using default model {args.model}, which is quite small and may not yield very accurate results; use --model to override it.')
    if not args.openai and 'gpt4' in args.model:
        logging.warning('WARNING: If you meant to use a model available through the OpenAI API, include --openai.')
    if args.openai and not prompt_template.is_chat:
        logging.warning('WARNING: Given --openai, you\'re advised to use a chat-style prompt.')
    if 'instruct' in args.model.lower() and not prompt_template.is_chat:
        logging.warning('WARNING: Given "instruct" model, you\'re advised to use a chat-style prompt.')

    model = MultipleChoiceModel(
        model_name=args.model,
        labels=prompt_template.labels_for_logits,
        prompt_start_for_cache=prompt_template.prompt_start_for_cache,
        openai_client=OpenAI(api_key=os.environ.get('OPENAI_API_KEY')) if args.openai else None
    )

    match prompt_template.mode:
        case 'comparative':
            if is_multicolumn:
                if args.compare_to:
                    logging.warning('WARNING: --compare_to is ignored, because input file is multi-column csv.')
                items = ({'choices': choices, 'prompt': prompt_template.format(*choices)} for choices in inputs)
            else:
                if args.compare_to:
                    compare_to, is_multicolumn2 = read_inputs(args.compare_to, args.csv, args.newlines)
                    if is_multicolumn2:
                        raise ValueError('The --compare_to file appears to contain multiple .csv columns, which is not allowed.')
                    compare_to = list(compare_to)
                else:
                    compare_to = inputs = list(inputs)
                items = iter_items_comparison(inputs, args.n_comparisons, args.all_positions, prompt_template,
                                              compare_to=compare_to, seed_per_item=args.seed if args.compare_deterministic else None)
            add_results_to_dict = add_results_comparative
        case 'categorical':
            items = iter_items_basic(inputs, prompt_template)
            add_results_to_dict = functools.partial(add_results_categorical, category_names=list(prompt_template.categories))
        case 'scalar':
            items = iter_items_basic(inputs, prompt_template)
            add_results_to_dict = functools.partial(add_results_scalar, scale=prompt_template.scale)

    logging.info(f'-------\n{prompt_template}\n-------')

    csv_writer = DictWriterAutoHeader(sys.stdout)
    for item in items:
        probs = model.get_probs(item['prompt'])
        add_results_to_dict(item, probs)
        del item['prompt']
        csv_writer.writerow(item)


def read_inputs(file, is_csv, escaped_newlines) -> tuple[Generator[Union[str, list[str]], None, None], bool]:
    is_multicolumn = False
    if is_csv:
        rows = csv.reader(file)
        first_item = next(rows)
        if len(first_item) > 1:
            is_multicolumn = True
            items = itertools.chain([first_item], rows)
        else:
            items = itertools.chain([first_item[0]], (l[0] for l in rows))
    elif escaped_newlines:
        items = (line.strip().replace('\\n', '\n') for line in file)
    else:
        items = (line.strip() for line in file)
    return items, is_multicolumn


def add_results_scalar(item: dict, probs: list[float], scale: list[int | float]):
    max_index = max(range(len(probs)), key=lambda x: probs[x])
    probs = [round(s, N_DECIMALS) for s in probs]
    item['pred'] = scale[max_index]
    item['prob'] = probs[max_index]
    item['rating'] = round(sum(s * n for s, n in zip(probs, scale)), N_DECIMALS)
    item['probs'] = make_csv_string(probs, delimiter=';')


def add_results_comparative(item: dict, probs: list[float]):
    max_index = max(range(len(probs)), key=lambda x: probs[x])
    probs = [round(s, N_DECIMALS) for s in probs]
    item['pred'] = item['choices'][max_index]
    if 'position' in item:
        item['prob'] = probs[item['position']]
    item['probs'] = make_csv_string(probs, delimiter=';')
    item['choices'] = make_csv_string(item['choices'], delimiter=';')


def add_results_categorical(item: dict, probs: list[float], category_names: list[str]):
    max_index = max(range(len(probs)), key=lambda x: probs[x])
    probs = [round(s, N_DECIMALS) for s in probs]
    item['pred'] = category_names[max_index]
    item['prob'] = probs[max_index]
    item['probs'] = make_csv_string(probs, delimiter=';')


def iter_items_basic(lines: Iterable[str], prompt_template: Union[str, PromptTemplate]) -> Generator[dict, None, None]:
    for n, line in enumerate(lines):
        prompt = prompt_template.format(line)
        yield {'target_id': n, 'target': line, 'prompt': prompt}


def iter_items_comparison(items: Iterable[str], n_comparisons: int, all_positions: bool,
                          prompt_template: Union[str, PromptTemplate], compare_to: list[str],
                          seed_per_item: int) -> Generator[dict, None, None]:

    n_choices = prompt_template.n_choices
    n_alternatives = n_choices - 1

    logging.info(f'Will do {n_comparisons * (n_choices if all_positions else 1)} comparisons per input line.')

    for item_id, item in enumerate(items):

        if seed_per_item is not None:
            random.seed(seed_per_item)

        all_alternatives = random_sample_not_containing(compare_to, n_comparisons * n_alternatives, item_to_exclude=item)

        for comp_id, alternatives in enumerate(batched(all_alternatives, n_alternatives)):
            positions = range(n_choices) if all_positions else [random.randint(0, n_alternatives)]
            for pos in positions:
                choices = alternatives[:pos] + [item] + alternatives[pos:]
                prompt = prompt_template.format(*choices)
                yield {'target_id': item_id, 'comparison_id': comp_id, 'position': pos, 'target': item, 'choices': choices, 'prompt': prompt}


def random_sample_not_containing(items: list, k: int, item_to_exclude) -> list:
    """
    Like random.sample(items, k), but excluding a specific element from the original. Original order not maintained.
    """
    sample = random.sample(items, k=min(k + 1, len(items)))  # one more if possible just in case item is among them
    try:
        sample.remove(item_to_exclude)
    except ValueError:  # if item not found
        if len(sample) > k:
            sample.pop()

    if len(sample) < k:
        raise ValueError(
            f'Not enough comparison items for n_choices × n_comparisons comparisons per item. '
            f'Decrease n_choices or --n_comparisons, or provide a longer list of items to compare '
            f'to (--compare_to).')

    return sample


def batched(iterable, n, *, strict=False) -> Generator[list, None, None]:
    """
    Was added to itertools only in Python 3.12, so included here.
    >>> batched('ABCDEFG', 3)
    ['A', 'B', 'C'], ['D', 'E', 'F'], ['G']
    """
    if n < 1:
        raise ValueError('n must be at least one')
    iterator = iter(iterable)
    while batch := list(itertools.islice(iterator, n)):
        if strict and len(batch) != n:
            raise ValueError('batched(): incomplete batch')
        yield batch


class DictWriterAutoHeader(csv.DictWriter):
    """
    A csv.DictWriter wrapper that lazily sets the fieldnames and writes the header automatically based
    on the first writerow call.
    """

    def __init__(self, f):
        super().__init__(f, fieldnames=[])

    def writerow(self, rowdict):
        if not self.fieldnames:
            self.fieldnames.extend(rowdict.keys())
            self.writeheader()
        super().writerow(rowdict)


def make_csv_string(row: list, delimiter=','):
    output = io.StringIO()
    writer = csv.writer(output, delimiter=delimiter)
    writer.writerow(row)
    csv_string = output.getvalue().strip()
    return csv_string


if __name__ == '__main__':
    main()
